
from rkstiff.etd import ETDAS,phi1,phi2,phi3
import numpy as np
from scipy.linalg import expm

class _ETD35_Diagonal:
    """
    ETD35 diagonal system strategy for ETD35 solver 
    """
    def __init__(self,linop,NLfunc,contourM,contourR,modecutoff):
        self.linop = linop
        self.NLfunc = NLfunc
        self.M = contourM
        self.R = contourR
        self.modecutoff = modecutoff
        
        N = linop.shape[0]
        self._EL14, self._EL12, self._EL34, self._EL = [np.zeros(N,dtype=np.complex128) for _ in range(4)]
        self._a21, self._a31, self._a32, self._a41, self._a43, self._a51,\
                self._a52,self._a54, self._a61, self._a62, self._a63,\
                self._a64, self._a65, self._a71, self._a73, self._a74,\
                self._a75, self._a76 = [np.zeros(N,dtype=np.complex128) for _ in range(18)]   
        self._NL1, self._NL2, self._NL3,self._NL4, self._NL5,\
                self._NL6 = [np.zeros(N,dtype=np.complex128) for _ in range(6)]
        self._k = np.zeros(N,dtype=np.complex128)
        self._err = np.zeros(N,dtype=np.complex128)
        
    def _updateCoeffs(self,h):
        z = h*self.linop
        self._updateCoeffsDiagonal(h,z)
        
    def _updateCoeffsDiagonal(self,h,z):   
        # diagonal system -> L is 1D array (of independent modes)
        self._EL14 = np.exp(z/4.0)
        self._EL12 = np.exp(z/2.0)
        self._EL34 = np.exp(3.0*z/4.0)
        self._EL = np.exp(z)

        smallmode_idx = np.abs(z) < self.modecutoff
        # compute big mode coeffs
        zb = z[~smallmode_idx] # z big
        phi1_14 = h*phi1(zb/4)
        phi2_14 = h*phi2(zb/4)
        phi1_12 = h*phi1(zb/2)
        phi2_12 = h*phi2(zb/2)
        phi1_34 = h*phi1(3*zb/4)
        phi2_34 = h*phi2(3*zb/4)
        phi1_1 = h*phi1(zb)
        phi2_1 = h*phi2(zb)
        phi3_1 = h*phi3(zb)

        self._a21[~smallmode_idx] = phi1_14/4.
        self._a31[~smallmode_idx] = (phi1_14-phi2_14/2.)/4.
        self._a32[~smallmode_idx] = phi2_14/8.
        self._a41[~smallmode_idx] = (phi1_12 - phi2_12)/2.
        self._a43[~smallmode_idx] = phi2_12/2.
        self._a51[~smallmode_idx] = 3.0*(phi1_34 - 3.*phi2_34/4.)/4.0
        self._a52[~smallmode_idx] = -3*phi1_34/8.
        self._a54[~smallmode_idx] = 9*phi2_34/16.
        self._a61[~smallmode_idx] = (-77*phi1_1+59*phi2_1)/42.
        self._a62[~smallmode_idx] = 8*phi1_1/7.
        self._a63[~smallmode_idx] = (111*phi1_1-87*phi2_1)/28.
        self._a65[~smallmode_idx] = (-47*phi1_1+143*phi2_1)/84.
        self._a71[~smallmode_idx] = 7*(257*phi1_1 - 497*phi2_1 + 270*phi3_1)/2700
        # Paper has error in a73/b3 phi2 coefficient (states this is 497 but it is actually 467)
        self._a73[~smallmode_idx] = (1097*phi1_1 - 467*phi2_1 - 150*phi3_1)/1350
        self._a74[~smallmode_idx] = 2*(-49*phi1_1 + 199*phi2_1 - 135*phi3_1)/225
        self._a75[~smallmode_idx] = (-313*phi1_1 + 883*phi2_1 - 90*phi3_1)/1350
        self._a76[~smallmode_idx] = (509*phi1_1 - 2129*phi2_1 + 1830*phi3_1)/2700
        
        # compute small mode coeffs
        zs = z[smallmode_idx] # z small
        r = self.R*np.exp(2j*np.pi*np.arange(0.5,self.M)/self.M)
        rr, zz = np.meshgrid(r,zs)
        Z = zz+rr
        
        phi1_14 = h*np.sum(phi1(Z/4),axis=1)/self.M
        phi2_14 = h*np.sum(phi2(Z/4),axis=1)/self.M
        phi1_12 = h*np.sum(phi1(Z/2),axis=1)/self.M
        phi2_12 = h*np.sum(phi2(Z/2),axis=1)/self.M
        phi1_34 = h*np.sum(phi1(3*Z/4),axis=1)/self.M
        phi2_34 = h*np.sum(phi2(3*Z/4),axis=1)/self.M
        phi1_1 = h*np.sum(phi1(Z),axis=1)/self.M
        phi2_1 = h*np.sum(phi2(Z),axis=1)/self.M
        phi3_1 = h*np.sum(phi3(Z),axis=1)/self.M
         
        self._a21[smallmode_idx] = phi1_14/4.
        self._a31[smallmode_idx] = (phi1_14-phi2_14/2.)/4.
        self._a32[smallmode_idx] = phi2_14/8.
        self._a41[smallmode_idx] = (phi1_12 - phi2_12)/2.
        self._a43[smallmode_idx] = phi2_12/2.
        self._a51[smallmode_idx] = 3.0*(phi1_34 - 3.*phi2_34/4.)/4.0
        self._a52[smallmode_idx] = -3*phi1_34/8.
        self._a54[smallmode_idx] = 9*phi2_34/16.
        self._a61[smallmode_idx] = (-77*phi1_1+59*phi2_1)/42.
        self._a62[smallmode_idx] = 8*phi1_1/7.
        self._a63[smallmode_idx] = (111*phi1_1-87*phi2_1)/28.
        self._a65[smallmode_idx] = (-47*phi1_1+143*phi2_1)/84.
        self._a71[smallmode_idx] = 7*(257*phi1_1 - 497*phi2_1 + 270*phi3_1)/2700
        # Paper has error in a73/b3 phi2 coefficient (states this is 497 but it is actually 467)
        self._a73[smallmode_idx] = (1097*phi1_1 - 467*phi2_1 - 150*phi3_1)/1350
        self._a74[smallmode_idx] = 2*(-49*phi1_1 + 199*phi2_1 - 135*phi3_1)/225
        self._a75[smallmode_idx] = (-313*phi1_1 + 883*phi2_1 - 90*phi3_1)/1350
        self._a76[smallmode_idx] = (509*phi1_1 - 2129*phi2_1 + 1830*phi3_1)/2700

    def _new_step_init(self,u):
        # Need to recompute NL1 if step was accepted (if not accepted use previously computed value)
        self._NL1 = self.NLfunc(u)
        
    def _updateStages(self,u,h,accept):        
        # One passthrough of RK solver for diagonal system 
        if accept:
            self._new_step_init(u)

        self._k = self._EL14*u + self._a21*self._NL1
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL14*u + self._a31*self._NL1 + self._a32*self._NL2
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL12*u + self._a41*self._NL1 + self._a43*self._NL3
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL34*u + self._a51*self._NL1 + self._a52*(self._NL2 - self._NL3) \
                               + self._a54*self._NL4
        self._NL5 = self.NLfunc(self._k)
        self._k = self._EL*u + self._a61*self._NL1 + self._a62*(self._NL2 - 3*self._NL4/2.0) \
                             + self._a63*self._NL3 + self._a65*self._NL5
        self._NL6 = self.NLfunc(self._k)
        self._k = self._EL*u + self._a71*self._NL1 + self._a73*self._NL3 + self._a74*self._NL4 \
                             + self._a75*self._NL5 + self._a76*self._NL6
        self._err = self._a75*(-self._NL1 + 4*self._NL3 - 6*self._NL4 + 4*self._NL5 - self._NL6)
        return self._k, self._err

class _ETD35_Diagonalized(_ETD35_Diagonal):
    """
    ETD35 non-diagonal system with eigenvector diagonalization strategy for ETD35 solver 
    """
    def __init__(self,linop,NLfunc,contourM,contourR,modecutoff):
        super().__init__(linop,NLfunc,contourM,contourR,modecutoff)
        if len(linop.shape) == 1:
            raise Exception('Cannot diagonalize a 1D system')
        linop_cond = np.linalg.cond(linop)
        if linop_cond > 1e16:
            raise Exception('Cannot diagonalize a non-invertible linear operator L')
        if linop_cond > 1000:
            print('''Warning: linear matrix array has a large condition number of {:.2f}, 
            method may be unstable'''.format(linop_cond))
        self._eig_vals, self._S = np.linalg.eig(linop)
        self._Sinv = np.linalg.inv(self._S)
        self._v = np.zeros(linop.shape[0])
        
    def _updateCoeffs(self,h):
        z = h*self._eig_vals
        self._updateCoeffsDiagonal(h,z)        
        
    def _new_step_init(self,u):
        self._NL1 = self._Sinv.dot(self.NLfunc(u))
        self._v = self._Sinv.dot(u)
    
    def _updateStages(self,u,h,accept):        
        # One passthrough of RK solver for diagonalized non-diagonal system 
        if accept:
            self._new_step_init(u)

        self._k = self._EL14*self._v + self._a21*self._NL1
        self._NL2 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._k = self._EL14*self._v + self._a31*self._NL1 + self._a32*self._NL2
        self._NL3 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._k = self._EL12*self._v + self._a41*self._NL1 + self._a43*self._NL3
        self._NL4 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._k = self._EL34*self._v + self._a51*self._NL1 + self._a52*(self._NL2 - self._NL3) \
                               + self._a54*self._NL4
        self._NL5 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._k = self._EL*self._v + self._a61*self._NL1 + self._a62*(self._NL2 - 3*self._NL4/2.0) \
                             + self._a63*self._NL3 + self._a65*self._NL5
        self._NL6 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._k = self._EL*self._v + self._a71*self._NL1 + self._a73*self._NL3 + self._a74*self._NL4 \
                             + self._a75*self._NL5 + self._a76*self._NL6
        self._err = self._a75*(-self._NL1 + 4*self._NL3 - 6*self._NL4 + 4*self._NL5 - self._NL6)
        return self._S.dot(self._k), self._err
        
class _ETD35_NonDiagonal:
    """
    ETD35 non-diagonal system strategy for ETD35 solver 
    """
    def __init__(self,linop,NLfunc,contourM,contourR):
        self.linop = linop
        self.NLfunc = NLfunc
        self.M = contourM
        self.R = contourR
        
        N = linop.shape[0]
        self._EL14, self._EL12, self._EL34, self._EL = [np.zeros(shape=linop.shape,dtype=np.complex128) for _ in range(4)]
        self._a21, self._a31, self._a32, self._a41, self._a43, self._a51,\
                self._a52,self._a54, self._a61, self._a62, self._a63,\
                self._a64, self._a65, self._a71, self._a73, self._a74,\
                self._a75, self._a76 = [np.zeros(shape=linop.shape,dtype=np.complex128) for _ in range(18)]   
        self._NL1, self._NL2, self._NL3,self._NL4, self._NL5,\
                self._NL6 = [np.zeros(N,dtype=np.complex128) for _ in range(6)]
        self._k = np.zeros(N,dtype=np.complex128)
        self._err = np.zeros(N,dtype=np.complex128)
        
    def _updateCoeffs(self,h):   
        # Update RK coefficients for 'matrix' np.array operator linop

        z = h*self.linop
        # Use expm matrix exponential function from scipy
        self._EL14 = expm(z/4.0)
        self._EL12 = expm(z/2.0)
        self._EL34 = expm(3.0*z/4.0)
        self._EL = expm(z)

        # Use contour integral evaluation for psi etd functions
        contour_points = self.R*np.exp(2j*np.pi*np.arange(0.5,self.M)/self.M)
        phi1_14,phi2_14,phi1_12,phi2_12,phi1_34,phi2_34 = [\
                np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(6)]
        phi1_1,phi2_1,phi3_1 = [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(3)]

        for point in contour_points:
            Q14 = np.linalg.inv(point*np.eye(*self.linop.shape)-z/4.)
            Q12 = np.linalg.inv(point*np.eye(*self.linop.shape)-z/2.)
            Q34 = np.linalg.inv(point*np.eye(*self.linop.shape)-3*z/4.)
            Q = np.linalg.inv(point*np.eye(*self.linop.shape)-z)
            phi1_14 += point*phi1(point)*Q14/self.M
            phi2_14 += point*phi2(point)*Q14/self.M
            phi1_12 += point*phi1(point)*Q12/self.M
            phi2_12 += point*phi2(point)*Q12/self.M
            phi1_34 += point*phi1(point)*Q34/self.M
            phi2_34 += point*phi2(point)*Q34/self.M
            phi1_1 += point*phi1(point)*Q/self.M
            phi2_1 += point*phi2(point)*Q/self.M
            phi3_1 += point*phi3(point)*Q/self.M  
        

        self._a21 = h*phi1_14/4.
        self._a31 = h*(phi1_14-phi2_14/2.)/4.
        self._a32 = h*phi2_14/8.
        self._a41 = h*(phi1_12 - phi2_12)/2.
        self._a43 = h*phi2_12/2.
        self._a51 = h*3.0*(phi1_34 - 3.*phi2_34/4.)/4.0
        self._a52 = -3*h*phi1_34/8.
        self._a54 = h*9*phi2_34/16.
        self._a61 = h*(-77*phi1_1+59*phi2_1)/42.
        self._a62 = h*8*phi1_1/7.
        self._a63 = h*(111*phi1_1-87*phi2_1)/28.
        self._a65 = h*(-47*phi1_1+143*phi2_1)/84.
        self._a71 = h*7*(257*phi1_1 - 497*phi2_1 + 270*phi3_1)/2700
        self._a73 = h*(1097*phi1_1 - 467*phi2_1 - 150*phi3_1)/1350
        self._a74 = h*2*(-49*phi1_1 + 199*phi2_1 - 135*phi3_1)/225
        self._a75 = h*(-313*phi1_1 + 883*phi2_1 - 90*phi3_1)/1350
        self._a76 = h*(509*phi1_1 - 2129*phi2_1 + 1830*phi3_1)/2700

    def _new_step_init(self,u):
        self._NL1 = self.NLfunc(u)
        
    def _updateStages(self,u,h,accept):        
        if accept:
            self._new_step_init(u)
            
        self._k = self._EL14.dot(u) + self._a21.dot(self._NL1)
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL14.dot(u) + self._a31.dot(self._NL1) + self._a32.dot(self._NL2)
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL12.dot(u) + self._a41.dot(self._NL1) + self._a43.dot(self._NL3)
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL34.dot(u) + self._a51.dot(self._NL1) + self._a52.dot(self._NL2 - self._NL3) \
                               + self._a54.dot(self._NL4)
        self._NL5 = self.NLfunc(self._k)
        self._k = self._EL.dot(u) + self._a61.dot(self._NL1) + self._a62.dot(self._NL2 - 3*self._NL4/2.0) \
                             + self._a63.dot(self._NL3) + self._a65.dot(self._NL5)
        self._NL6 = self.NLfunc(self._k)
        self._k = self._EL.dot(u) + self._a71.dot(self._NL1) + self._a73.dot(self._NL3) + self._a74.dot(self._NL4) \
                             + self._a75.dot(self._NL5) + self._a76.dot(self._NL6)
        self._err = self._a75.dot(-self._NL1 + 4*self._NL3 - 6*self._NL4 + 4*self._NL5 - self._NL6)
        return self._k, self._err


class ETD35(ETDAS):
    """
    Exponential time-differencing adaptive step solver of 5th order with 3rd order embedding

    ATTRIBUTES
    __________

    linop : np.array
    NLfunc : function
    t : time-array stored with evolve function call 
    u : output-array stored with evolve function call 
    logs : array of info stored related to the solver

    ETD Parameters (see ETDAS class in etd module)
    ______________
    modecutoff : float
    contour_points :  int
    contour_radius : float 

    StiffSolverAS Parameters (see StiffSolverAS class in solver module)
    ________________________
    epsilon : float
    incrF : float
    decrF : float 
    safetyF : float
    adapt_cutoff : float
    minh : float 
    """

    def __init__(self,linop,NLfunc,**kwargs):
        """
        INPUTS
        ______

        linop : np.array
            Linear operator (L) in the system dtU = LU + NL(U). Can be either a 2D numpy array (matrix)
            or a 1D array (diagonal system). L can be either real-valued or complex-valued.

        NLfunc : function 
            Nonlinear function (NL(U)) in the system dtU = LU + NL(U). Can be a complex or real-valued function.

        diagonalize : bool, optional
            Diagonalize the linear operator (matrix) and solve the diagonalized system 

        ETDAS variables: modecutoff, contour_points, contour_radius (see ETDAS documentation from etd module)

        StiffSolverAS variables: epsilon, incrF, decrF, safetyF, adapt_cutoff, minh 
                        (see StiffSolverAS documentation from solver module)
        """

        super().__init__(linop,NLfunc,**kwargs)
        self._method = None
        if self._diag:
            self._method = _ETD35_Diagonal(linop,NLfunc,self.contour_points,\
                    self.contour_radius,self.modecutoff)
        else:
            if 'diagonalize' in kwargs and kwargs['diagonalize']:
                self._method = _ETD35_Diagonalized(linop,NLfunc,self.contour_points,\
                        self.contour_radius,self.modecutoff)
            else:
                self._method = _ETD35_NonDiagonal(linop,NLfunc,self.contour_points,\
                        self.contour_radius)
        self._stages_init = False
        self._accept = False

    def _reset(self):
        #Resets solver to its initial state 
        self._stages_init = False
        self._h_coeff = None
        self._accept = False
        
    def _updateCoeffs(self,h):
        # Update coefficients if step size h changed
        if h == self._h_coeff:
            return
        self._h_coeff = h
        self._method._updateCoeffs(h)
        self.logs.append("ETD35 coefficients updated")
        
    def _updateStages(self,u,h):
        # Computes u_{n+1} from u_{n} through one RK passthrough
        self._updateCoeffs(h)
        if not self._stages_init:
            self._method._new_step_init(u)
            self._stages_init = True
        return self._method._updateStages(u,h,self._accept)
    
    def _q(self):
        # Order variable for computing suggested step size (embedded order + 1)
        return 4        







from rkstiff.solver import StiffSolverAS
import numpy as np
from scipy.linalg import expm

class _IF34_Diagonal:
    """
    IF34 diagonal system strategy for IF34 solver 
    """
    def __init__(self,linop,NLfunc):
        self.linop = linop
        self.NLfunc = NLfunc
        
        N = linop.shape[0]
        self._EL, self._EL2 = [np.zeros(N,dtype=np.complex128) for _ in range(2)]
        self._NL1, self._NL2, self._NL3,self._NL4, self._NL5 = [np.zeros(N,dtype=np.complex128) for _ in range(5)]
        self._k = np.zeros(N,dtype=np.complex128) 
        self._err = np.zeros(N,dtype=np.complex128)
        
    def _updateCoeffs(self,h):
        z = h*self.linop
        self._updateCoeffsDiagonal(h,z)
        
    def _updateCoeffsDiagonal(self,h,z):   
        self._EL = np.exp(z)
        self._EL2 = np.exp(z/2)
    
    def _N1_init(self,u):
        self._NL1 = self.NLfunc(u)
        
    def _updateStages(self,u,h,accept):        
        if accept:
            self._NL1 = self._NL5.copy()
        
        self._k = self._EL2*u + h*self._EL2*self._NL1/2.0
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL2*u + h*self._NL2/2.
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL*u + h*self._EL2*self._NL3
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL*u + h*(self._EL*self._NL1/6.0 + self._EL2*self._NL2/3.0 \
                + self._EL2*self._NL3/3.0 + self._NL4/6.0)
        self._NL5 = self.NLfunc(self._k)
        self._err = h*(self._NL4-self._NL5)/6.
        return self._k, self._err

class _IF34_Diagonalized(_IF34_Diagonal):
    """
    IF34 non-diagonal system with eigenvector diagonalization strategy for IF34 solver 
    """
    def __init__(self,linop,NLfunc):
        super().__init__(linop,NLfunc)
        if len(linop.shape) == 1:
            raise Exception('Cannot diagonalize a 1D system')
        linop_cond = np.linalg.cond(linop)
        if linop_cond > 1e16:
            raise Exception('Cannot diagonalize a non-invertible linear operator L')
        if linop_cond > 1000:
            print('''Warning: linear matrix array has a large condition number of {:.2f}, 
            method may be unstable'''.format(linop_cond))
        self._eig_vals, self._S = np.linalg.eig(linop)
        self._Sinv = np.linalg.inv(self._S)
        self._v = np.zeros(linop.shape[0])
        
    def _updateCoeffs(self,h):
        z = h*self._eig_vals
        self._updateCoeffsDiagonal(h,z)        
        
    def _N1_init(self,u):
        self._NL1 = self._Sinv.dot(self.NLfunc(u))
        self._v = self._Sinv.dot(u)
    
    def _updateStages(self,u,h,accept):        
        # Use First is same as last principle (FSAL) -> k5 stage is input u for next step
        if accept:
            self._NL1 = self._NL5.copy()    
            self._v = self._Sinv.dot(u)

        self._k = self._EL2*self._v + h*self._EL2*self._NL1/2.0
        self._NL2 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._k = self._EL2*self._v + h*self._NL2/2.
        self._NL3 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._k = self._EL*self._v + h*self._EL2*self._NL3
        self._NL4 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._k = self._EL*self._v + h*(self._EL*self._NL1/6.0 + self._EL2*self._NL2/3.0 \
                + self._EL2*self._NL3/3.0 + self._NL4/6.0)
        self._NL5 = self._Sinv.dot(self.NLfunc(self._S.dot(self._k)))
        self._err = h*(self._NL4-self._NL5)/6.
        return self._S.dot(self._k), self._err

class _IF34_NonDiagonal:
    """
    IF34 non-diagonal system strategy for IF34 solver 
    """
    def __init__(self,linop,NLfunc):
        self.linop = linop
        self.NLfunc = NLfunc
        
        N = linop.shape[0]
        self._EL, self._EL2 = [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(2)]
        self._NL1, self._NL2, self._NL3,self._NL4, self._NL5 = [np.zeros(N,dtype=np.complex128) for _ in range(5)]
        self._k = np.zeros(N,dtype=np.complex128) 
        self._err = np.zeros(N,dtype=np.complex128)
        
    def _updateCoeffs(self,h):   
        z = h*self.linop
        self._EL = expm(z)
        self._EL2 = expm(z/2)
        
    def _N1_init(self,u):
        self._NL1 = self.NLfunc(u)
        
    def _updateStages(self,u,h,accept):        
        # Use First is same as last principle (FSAL) -> k5 stage is input u for next step
        if accept:
            self._NL1 = self._NL5.copy()    

        self._k = self._EL2.dot(u) + h*self._EL2.dot(self._NL1/2.0)
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL2.dot(u) + h*self._NL2/2.
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL.dot(u) + h*self._EL2.dot(self._NL3)
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL.dot(u) + h*(self._EL.dot(self._NL1/6.0) + self._EL2.dot(self._NL2/3.0) \
                + self._EL2.dot(self._NL3/3.0) + self._NL4/6.0)
        self._NL5 = self.NLfunc(self._k)
        self._err = h*(self._NL4-self._NL5)/6.
        return self._k, self._err

class IF34(StiffSolverAS):
    """
    Integrating factor adaptive step solver of 4th order with 3rd order embedding

    ATTRIBUTES
    __________
    linop : np.array
    NLfunc : function
    t : time-array stored with evolve function call 
    u : output-array stored with evolve function call 
    logs : array of info stored related to the solver

    StiffSolverAS Parameters (see StiffSolverAS class in solver module)
    ________________________
    epsilon : float
    incrF : float
    decrF : float 
    safetyF : float
    adapt_cutoff : float
    minh : float 
    """

    def __init__(self,linop,NLfunc,**kwargs):
        """
        INPUTS
        ______

        linop : np.array
            Linear operator (L) in the system dtU = LU + NL(U). Can be either a 2D numpy array (matrix)
            or a 1D array (diagonal system). L can be either real-valued or complex-valued.

        NLfunc : function 
            Nonlinear function (NL(U)) in the system dtU = LU + NL(U). Can be a complex or real-valued function.

        diagonalize : bool, optional
            Diagonalize the linear operator (matrix) and solve the diagonalized system 

        StiffSolverAS variables: epsilon, incrF, decrF, safetyF, adapt_cutoff, minh 
                        (see StiffSolverAS documentation from solver module)
        """

        super().__init__(linop,NLfunc,**kwargs)
        self._method = None
        if self._diag:
            self._method = _IF34_Diagonal(linop,NLfunc)
        else:
            if 'diagonalize' in kwargs and kwargs['diagonalize']:
                self._method = _IF34_Diagonalized(linop,NLfunc)
            else:
                self._method = _IF34_NonDiagonal(linop,NLfunc)
        self._reset()

    def _reset(self):
        #Resets solver to its initial state 
        self.__N1_init = False
        self._h_coeff = None
        self._accept = False
        
    def _updateCoeffs(self,h):
        # Update coefficients if step size h changed
        if h == self._h_coeff:
            return
        self._h_coeff = h
        self._method._updateCoeffs(h)
        self.logs.append("IF34 coefficients updated")
        
    def _updateStages(self,u,h):
        # Computes u_{n+1} from u_{n} through one RK passthrough
        self._updateCoeffs(h)
        if not self.__N1_init:
            self._method._N1_init(u)
            self.__N1_init = True
        return self._method._updateStages(u,h,self._accept)
    
    def _q(self):
        # Order variable for computing suggested step size (embedded order + 1)
        return 4        






import numpy as np

def kdvSoliton(x,A=0.5,x0=0,t=0):
    u0 = 0.5*A**2/(np.cosh(A*(x-x0-A**2*t)/2)**2)
    return u0

def kdvMultiSoliton(x,A,x0,t=0):
    assert len(x0) == len(A)
    M = len(A)
    N = len(x)
    A = np.array(A).reshape(1,M)
    x0 = np.array(x0).reshape(1,M)
    u0 = 0.5*A**2/(np.cosh(A*(x.reshape(N,1)-x0-A**2*t)/2)**2)
    u0 = np.sum(u0,axis=1)
    return u0

def kdvOps(kx):
    L = 1j*kx**3
    def NL(uf):
        u = np.fft.irfft(uf)
        ux = np.fft.irfft(1j*kx*uf)
        return -6*np.fft.rfft(u*ux)
    return L,NL


def burgersOps(kx,mu):
    L = -mu*kx**2
    def NL(uf):
        u = np.fft.irfft(uf)
        ux = np.fft.irfft(1j*kx*uf)
        return -np.fft.rfft(u*ux)
    return L,NL

def allenCahnOps(x,Dx,epsilon):
    epsilon = 0.01
    D2 = Dx.dot(Dx)
    L = epsilon*D2 + np.eye(*D2.shape)
    L = L[1:-1,1:-1] # remove boundary points
    def NL(u):
        return x[1:-1] - np.power(u+x[1:-1],3)
    return L,NL











import numpy as np
import scipy.special as sp

def construct_x_kx_rfft(N,a=0.0,b=2*np.pi):
    """ Constructs a uniform 1D spatial grid and rfft spectral wavenumbers for real-valued functions
    INPUTS
        N - even integer greater than 2
        a - left endpoint in spatial grid
        b - right endpoint in spatial grid
    OUTPUTS
        x - uniform 1D spatial grid
        kx - spectral wavenumber grid
    """

    if not isinstance(N,int):
        raise TypeError('Number of grid points N must be an integer, it is {}'.format(N))
    if N <= 2:
        raise ValueError('Number of grid points N must be larger than 2, it is {}'.format(N))
    if (N % 2) != 0:
        raise ValueError('Integer N in construct_x_kx_rfft must be an even number')

    dx = (b-a)/N
    x = np.arange(a,b,dx)
    kx = 2*np.pi*np.fft.rfftfreq(N,d=dx)
    return x,kx


def construct_x_kx_fft(N,a=0.0,b=2*np.pi):
    """ Constructs a uniform 1D spatial grid and fft spectral wavenumbers for complex-valued functions
    INPUTS
        N - even integer greater than 2
        a - left endpoint in spatial grid
        b - right endpoint in spatial grid
    OUTPUTS
        x - uniform 1D spatial grid
        kx - spectral wavenumber grid
    """

    if not isinstance(N,int):
        raise TypeError('Number of grid points N must be an integer, it is {}'.format(N))
    if N <= 2:
        raise ValueError('Number of grid points N must be larger than 2, it is {}'.format(N))
    if (N % 2) != 0:
        raise ValueError('Integer N in construct_x_kx_rfft must be an even number')

    dx = (b-a)/N
    x = np.arange(a,b,dx)
    kx = 2*np.pi*np.fft.fftfreq(N,d=dx)
    return x,kx

def construct_x_cheb(N,a=-1,b=1):
    """ Constructs a 1D grid with Chebyshev spatial discretization 

    INPUTS
        N - positive integer
        a - left endpoint in spatial grid
        b - right endpoint in spatial grid
    OUTPUTS
        x - grid discretized at Chebyshev points
    """

    if not isinstance(N,int):
        raise TypeError('Max Chebyshev grid point number N must be an integer, it is {}'.format(N))
    if N < 2:
        raise ValueError('Max Chebyshev grid point number N must be larger than 1, it is {}'.format(N))
    x = np.polynomial.chebyshev.chebpts2(N+1)
    x = a+(b-a)*(x+1)/2.
    return x

def construct_x_Dx_cheb(N,a=-1,b=1):
    """ Constructs a 1D grid with Chebyshev spatial discretization along with a
        differentiation matrix for functions sampled on this grid

    INPUTS
        N - positive integer
        a - left endpoint in spatial grid
        b - right endpoint in spatial grid
    OUTPUTS
        x - grid discretized at Chebyshev points
        Dx - derivative matrix for functions sampled at Chebyshev points
    """
    x = construct_x_cheb(N,a,b)
    c = np.r_[2,np.ones(N-1),2]*np.power(-1,np.arange(0,N+1))
    X = np.tile(x.reshape(N+1,1),(1,N+1)) # put copies of x in columns (first row is x0)
    dX = X - X.T
    Dx = np.outer(c,1./c)/(dX+np.eye(N+1))
    Dx = Dx - np.diag(Dx.sum(axis=1))
    return x,Dx


class HankelTransform:
    """
    A class for computing discrete Hankel Transforms

    ATTRIBUTES
    __________

    nr : int
        Number of radial points sampled. The size of the
        hankel transform matrix is nr x nr.

    rmax : float
        Maximum radius of sampled points. 

    r : np.array, dtype = float
        Radial points for a spectral grid suitable for the Hankel transform.
        This grid is determined by the user specification of nr and rmax

    kr : np.array, dtype = float
        Spectral points for a radial grid suitable for the Hankel transform.
        This grid is determined by the user specification of nr and rmax

    METHODS
    _______

    ht(f):
        Computes a hankel transform of the function f sampled at the radial points 
        specified by r.

    iht(g):
        Computes an inverse hankel transform of the spectral space function g sampled
        at the spectral points specified by kr

    """

    def __init__(self,nr,rmax=1.0):
        """
        Constructs a Hankel transform Matrix that is used in the forward hankel transform
        function ht and inverse hankel transform function iht

        INPUTS
        ______

        nr : int
            Number of radial points sampled. The size of the
            hankel transform matrix is nr x nr. nr >= 4

        rmax :  float
            Maximum radius of sampled points. rmax > 0

        OUTPUTS
        _______
        None

        """

        self._bessel_zeros = None
        self._jN = None
        self._Y = None
        self._setnr_setrmax(nr,rmax)

    def _setnr_setrmax(self,nr,rmax):
        self._setbesselzeros(nr)
        self._setgrids(rmax)
        self._sethankelmatrix()
        self._rmax = rmax
        self._nr = nr 

    def _setbesselzeros(self,nr):
        # find and save first nr bessel zeros in an array, also save the nr+1 bessel zero
        bessel_zeros = sp.jn_zeros(0,nr+1)
        self._bessel_zeros,self._jN = bessel_zeros[:-1], bessel_zeros[-1]

    def _setgrids(self,rmax):
        # set the r and kr radial grids given a maximum radius rmax
        self.r = self._bessel_zeros*rmax/self._jN
        self.kr = self._bessel_zeros/rmax

    def _sethankelmatrix(self):
        # set the Hankel matrix used in the Hankel transform for the saved radial grid
        j1vec = sp.j1(self._bessel_zeros)
        bessel_arg = np.outer(self._bessel_zeros,self._bessel_zeros)/self._jN
        self._Y = 2*sp.j0(bessel_arg)/(self._jN*j1vec**2)

    def _scalefactor(self):
        # factor used in transforming from real space to spectral space and vice_versa for internal use
        return self.rmax**2/self._jN

    def hankelMatrix(self):
        """ Returns a copy of the Hankel transform matrix constructed by the class. """
        return self._Y.copy()

    def besselZeros(self):
        """ Returns a copy of the Bessel zeros used by the Hankel transform. """
        return self._bessel_zeros.copy()

    @property
    def nr(self):
        """ Returns the number of radial points for the radial grid specified by the class. """
        return self._nr

    @nr.setter
    def nr(self,nr):
        """ Sets the number of radial points in the grid to be used by the Hankel transform. """
        if not isinstance(nr,int):
            raise ValueError('nr must be an integer')
        if nr < 4: 
            raise ValueError('nr must be greater than or equal to 4')
        self._setbesselzeros(nr)
        self._setgrids(self.rmax)
        self._sethankelmatrix()
        self._nr = nr 

    @property
    def rmax(self):
        """ Returns the maximum radius of the radial grid used by the Hankel transform """
        return self._rmax

    @rmax.setter
    def rmax(self,rmax):
        """ Sets the maximum radius of the radial grid used by the Hankel transform """
        if rmax <= 0:
            raise ValueError('rmax must be non-negative')
        self._setgrids(rmax)
        self._rmax = rmax 

    def ht(self,f):
        """
        Computes a hankel transform of the function f sampled at the radial
        points specified by r
        
        INPUTS
        ______
        
        f : np.array, dtype=float
            function sampled at the discretized points specified by r

        OUTPUTS
        _______

        g :  np.array, dtype=float
            Hankel spectral space representation of the function f corresponding
            to the spectral space grid points kr
        """

        return self._scalefactor()*self._Y.dot(f)

    def iht(self,g):
        """
        Computes an inverse hankel transform of the function g sampled at the spectral
        space points specified by kr
        
        INPUTS
        ______
        
        g : np.array, dtype=float
            spectral space function sampled at the discretized points specified by kr

        OUTPUTS
        _______

        f :  np.array, dtype=float
            Real-space representation of the spectral space function g corresponding
            to values on the radial grid r
        """

        return self._Y.dot(g)/self._scalefactor()


def mirrorGrid(r,u=None,axis=-1):
    """
    Converts r grid from [0,rmax] interval to [-rmax,rmax] interval and adjusts
    function output u accordingly

    INPUTS
    ______

    r : np.array, dytpe=float
        radial grid on interval [0,rmax]

    u : np.array
        function values specified at radial points given by r

    axis : int 
        axis value determines how to mirror the u array (-1 -> stack horizontally,
        0 -> stack vertically)

    OUTPUTS
    _______

    rnew : np.array, dtype=float
        'radial' grid on the interval [-rmax,rmax] 

    unew : np.array
        function values specified at 'radial' points given by rnew

    """

    rnew = np.hstack([-np.flipud(r),r])
    if u is None:
        return rnew

    if axis == -1:
        unew = np.hstack([np.flipud(u),u])
    elif axis == 0:
        unew = np.vstack([np.flipud(u),u])
    elif axis == 1:
        unew = np.hstack([np.fliplr(u),u])
    else:
        raise ValueError('axis variable must be -1 or 0')

    return rnew,unew






import numpy as np
from rkstiff.solver import StiffSolverAS

def phi1(z):
    """ Computes RKETD psi-function of the first order. 
    INPUTS
        z - real or complex-valued input array 
    OUTPUT
        return (exp(z)-1)/z  -> real or complex-valued RKETD function of first order
    """
    return (np.exp(z)-1)/z

def phi2(z):
    """ Computes RKETD psi-function of the second order. 
    INPUTS
        z - real or complex-valued input array 
    OUTPUT
        return 2!*(exp(z)-1-z/2)/z^2  -> real or complex-valued RKETD function of second order
    """
    return 2*(np.exp(z)-1-z)/z**2

def phi3(z):
    """ Computes RKETD psi-function of the third order. 
    INPUTS
        z - real or complex-valued input array 
    OUTPUT
        return 3!*(exp(z)-1-z/2-z^3/6)/z^3  -> real or complex-valued RKETD function of third order
    """
    return 6*(np.exp(z)-1-z-z**2/2)/z**3

class ETDAS(StiffSolverAS):
    """
    Class template for an RKETD adaptive-step solver. Adds several ETD specific parameters
    to the StiffSolverAS class.

    ATTRIBUTES
    ----------

    modecutoff : float
       psi RKETD functions are numerically inaccurate near zero when computed directly; when |z| < modecutoff use a
       contour integral (more computationally expensive method) to compute these values
    contour_points : int
        number of points to use in contour integral when evaluating small input values to the psi RKETD functions
    contour_radius : float
        radius size of a circular contour integral to be used to evaluate small input values to the psi RKETD functions

    """

    def __init__(self,linop,NLfunc,**kwargs):
        super().__init__(linop,NLfunc,**kwargs)
        self.modecutoff = 0.01
        if 'modecutoff' in kwargs:
            self.modecutoff = kwargs['modecutoff']
            if (self.modecutoff > 1) or (self.modecutoff <= 0):
                raise ValueError('modecutoff must be between 0 and 1 but is {}'.format(self.modecutoff))
        self.contour_points = 32
        if 'contour_points' in kwargs:
            self.contour_points = kwargs['contour_points']
            if not isinstance(self.contour_points,int):
                raise TypeError('contour_points must be an integer but is {}'.format(self.contour_points))
            if self.contour_points <= 1:
                raise ValueError('contour_points must be an integer greater than 1 but is {}'.format(self.contour_points))
        self.contour_radius = 1.0
        if 'contour_radius' in kwargs:
            self.contour_radius = kwargs['contour_radius']
            if self.contour_radius <= 0:
                raise ValueError('contour_radius must greater than 0 but is {}'.format(self.contour_radius))
        self._h_coeff = None
        










import numpy as np

class StiffSolverAS:
    """
    Base class for an adaptive-step Runge-Kutta solver for stiff systems of the type dtU = LU + NL(U),
    where L is a linear operator and NL is a non-linear function.

    ATTRIBUTES
    __________

    linop : np.array
        Linear operator (L) in the system dtU = LU + NL(U). Can be either a 2D numpy array (matrix)
        or a 1D array (diagonal system). L can be either real-valued or complex-valued.

    NLfunc : function 
        Nonlinear function (NL(U)) in the system dtU = LU + NL(U). Can be a complex or real-valued function.

    u : list
        List of np.arrays corresponding to the propagated U in the system dtU = LU + NL(U) using the evolve function

    t : list
        List of times corresponding to the propagated U in the system dtU = LU + NL(U) using the evolve function

    logs : list
        List of log messages

    incrF : float, > 1.0
        Increment factor for increasing the step size utilized in propagating a system. After each step in the propagation
        a 'optimal' step size of h_opt is computed and compared vs the current step size h_current. If h_opt > incrF*h_current,
        then the solver will suggest h_opt as the next step size, otherwise it will continue using the current step size. This
        parameter is used such that very small changes to the step size are avoided and hence potentially expensive evaluations
        of the RK coefficients are avoided. 

    decrF : float, < 1.0
        Decrement factor for decreasing the step size utilized in propagating a system. After each step in the propagation
        a 'optimal' step size of h_opt is computed and compared vs the current step size h_current. If h_opt > decrF*h_current and
        h_opt < h_current, then the solver will suggest h_new = decrF*h_current as the next step size. This
        parameter is used such that very small changes to the step size are avoided and hence potentially expensive evaluations
        of the RK coefficients are avoided. 

    epsilon : float
        Relative error tolerance for system. Solver will suggest step sizes in an attempt to keep the two-norm relative error of the system
        less than this value. This is used as a tuning parameter in the solver and the relative-error is not strictly
        enforced! In general, smaller epsilon results in smaller relative-error but due to the nature of the solvers utilized, the
        adaptive stepping is prone to be less than ideal and errors can often be larger than the specified epsilon tolerance level.

    safetyF : float
        Safety factor for adaptive stepping. The 'optimal' step size computed using the embedded RK methods is multiplied by
        this factor to try to enforce the relative-error to be below the epsilon relative-error threshold. The relative-error
        computed for these solvers based on embedded methods is an inaccurate estimate and so this saftey factor can be used
        to tune the system error to be more inline with actual errors.

    adapt_cutoff : float < 1
        Limits values used in the computation of the suggested step size to those with |u| > adapt_cutoff*max(|u|). To include all
        values in the calculation of step size: set this to a very small number. 

    minh : float
        Minimum step size that can be taken by the solver before throwing an exception


    METHODS
    _______

    step(u,h_suggest):
        Propagates a given array of u one step using an RK method for stiff PDEs. 

    evolve(u,t0,tf,h_init=None,**kwargs)
        Propagates an initial value (array) of u given at time t0
        until a final time tf is reached using a RK method for stiff PDEs 

    reset()
        Resets solver including erasing stored variables such as self.t, self.u, and self.logs

    """
    MAX_LOOPS = 50
    MAX_S = 4
    MIN_S = 0.25
    
    def __init__(self,linop,NLfunc,**kwargs):
        """

        linop : np.array
            Linear operator (L) in the system dtU = LU + NL(U). Can be either a 2D numpy array (matrix)
            or a 1D array (diagonal system). L can be either real-valued or complex-valued.

        NLfunc : function 
            Nonlinear function (NL(U)) in the system dtU = LU + NL(U). Can be a complex or real-valued function.

        incrF : float, > 1.0
            Increment factor for increasing the step size utilized in propagating a system. After each step in the propagation
            a 'optimal' step size of h_opt is computed and compared vs the current step size h_current. If h_opt > incrF*h_current,
            then the solver will suggest h_opt as the next step size, otherwise it will continue using the current step size. This
            parameter is used such that very small changes to the step size are avoided and hence potentially expensive evaluations
            of the RK coefficients are avoided. 

        decrF : float, < 1.0
            Decrement factor for decreasing the step size utilized in propagating a system. After each step in the propagation
            a 'optimal' step size of h_opt is computed and compared vs the current step size h_current. If h_opt > decrF*h_current and
            h_opt < h_current, then the solver will suggest h_new = decrF*h_current as the next step size. This
            parameter is used such that very small changes to the step size are avoided and hence potentially expensive evaluations
            of the RK coefficients are avoided. 

        epsilon : float
            Relative error tolerance for system. Solver will suggest step sizes in an attempt to keep the two-norm relative error of the system
            less than this value. This is used as a tuning parameter in the solver and the relative-error is not strictly
            enforced! In general, smaller epsilon results in smaller relative-error but due to the nature of the solvers utilized, the
            adaptive stepping is prone to be less than ideal and errors can often be larger than the specified epsilon tolerance level.

        safetyF : float
            Safety factor for adaptive stepping. The 'optimal' step size computed using the embedded RK methods is multiplied by
            this factor to try to enforce the relative-error to be below the epsilon relative-error threshold. The relative-error
            computed for these solvers based on embedded methods is an inaccurate estimate and so this saftey factor can be used
            to tune the system error to be more inline with actual errors.

        adapt_cutoff : float < 1
            Limits values used in the computation of the suggested step size to those with |u| > adapt_cutoff*max(|u|). To include all
            values in the calculation of step size: set this to a very small number. 

        minh : float
            Minimum step size that can be taken by the solver before throwing an exception
            self.linop = linop
        """

        self.linop = linop
        self.NLfunc = NLfunc
        self.t, self.u = [], []

        dims = linop.shape
        self._diag = True
        if len(dims) > 2:
            raise ValueError('linop must be a 1D or 2D array')
        elif len(dims) == 2:
            if (dims[0] != dims[1]):
                raise ValueError('linop must be a square matrix')
            self._diag = False

        self.incrF = 1.25
        if 'incrF' in kwargs:
            self.incrF = kwargs['incrF']
            if self.incrF <= 1:
                raise ValueError('incrF must be > 1 but is {}'.format(self.incrF))
            
        self.decrF = 0.85
        if 'decrF' in kwargs:
            self.decrF = kwargs['decrF']
            if self.decrF >= 1:
                raise ValueError('decrF must be < 1 but is {}'.format(self.decrF))
            
        self.epsilon = 1.0e-4
        if 'epsilon' in kwargs:
            self.epsilon = kwargs['epsilon']
            if self.epsilon < 0:
                raise ValueError('epsilon must be positive but is {}'.format(self.epsilon))

        self.safetyF = 0.8
        if 'safetyF' in kwargs:
            self.safetyF = kwargs['safetyF']
            if self.safetyF > 1:
                raise ValueError('safetyF must be <= 1 but is {}'.format(self.safetyF))

        self.adapt_cutoff = 0.01
        if 'adapt_cutoff' in kwargs:
            self.adapt_cutoff = kwargs['adapt_cutoff']
            if self.adapt_cutoff >= 1:
                raise ValueError('adapt_cutoff must be < 1 but is {}'.format(self.adapt_cutoff))

        self.minh = 1.0e-16
        if 'minh' in kwargs:
            self.minh = kwargs['minh']
            if self.minh <= 0:
                raise ValueError('minh must be positive but is {}'.format(self.minh))

        self.__t0, self.__tf, self.__tc = 0,0,0
        self.__h_prev = 0.0
        self._accept = False 
        self.t, self.u, self.logs = [], [], []

    def reset(self):
        """ Resets solver to its initial state (such that its ready to call the functions evolve 
            or step on a new input). Erases stored variables such as self.t, self.u, and self.logs.
        """
        self.t, self.u, self.logs = [], [], []
        self.__t0, self.__tf, self.__tc = 0,0,0
        self.__h_prev = 0.0
        self._accept = False 
        self._reset()

    # reset solver dependent variables of the subclass
    def _reset():
        raise NotImplementedError
    
    # update RK stages, returns u_{n+1} given u_{n}, to be overwritten by subclass
    def _updateStages(self,u,h):
        raise NotImplementedError
        
    # q value in computation of suggested step size, to be overwritten by subclass
    def _q(self):
        raise NotImplementedError
        
    def step(self,u,h_suggest):
        """ 
        Propagates a given array of u one step using an RK method for stiff PDEs. 
        
        INPUTS:
            u : np.array
                input to propagate
            h_suggest : float
                suggested step size; may be reduced to achieve the desired accuracy as
                specified by the relative error threshold variable 'epsilon'

        OUTPUTS:
            unew : np.array
                result of input u propagated one step in the RK algorithm
            h : float
                actual step size taken in the algorithm (may be less than h_suggest)
            h_suggest : float
                step size the algorithm suggests you take for the next step
        """
        
        h = h_suggest
        assert h >= 0.0
        numloops = 0
        while(True):
            unew,err = self._updateStages(u,h)
            self.__h_prev = h 
            # Compute step size change factor s
            s = self._computeS(unew,err) 
            # If s is less than 1, inf, or nan, reject step and reduce step size
            if np.isinf(s) or np.isnan(s) or s < 1.0:      
                h = self._rejectStepSize(s,h)
            # If s is bigger than 1 accept h and the step
            else:
                h_suggest = self._acceptStepSize(s,h) 
                return unew, h, h_suggest 
            numloops += 1
            if numloops > self.MAX_LOOPS:
                failure_str = """Solver failed: adaptive step made too many attempts to find a step
                                 size with an acceptible amount of error. """
                self.logs.append(failure_str)
                raise Exception(failure_str)
            if h < self.minh:
                failure_str = """Solver failed: adaptive step reached minimum step size """
                self.logs.append(failure_str)
                raise Exception(failure_str)
                
        raise Exception("Propagation Failed")
        
    # helper function for computing coefficient s used in generating suggested step size
    def _computeS(self,u,err):
        # Use adapt_cutoff to ignore small modes/values in the computation of the step size
        magu = np.abs(u)
        idx = magu/magu.max() > self.adapt_cutoff
        tol = self.epsilon*np.linalg.norm(u[idx])
        s = self.safetyF*np.power(tol/np.linalg.norm(err[idx]),1.0/self._q())
        return s
    
    # helper function computing suggested step size if step is rejected
    def _rejectStepSize(self,s,h):
        self._accept = False
        # Check that s is a number
        if np.isinf(s) or np.isnan(s):
            self.logs.append('inf or nan number encountered: reducing step size to {}!'.format(h))
            return self.MIN_S*h
        
        s = np.max([s,self.MIN_S]) # dont let s be too small
        s = np.min([s,self.decrF]) # dont let s be too close to 1
        self.logs.append('step rejected with s = {:.2f}'.format(s))
        hnew = s*h
        self.logs.append('reducing step size to {}'.format(hnew))
        return hnew
    
    # helper function computing suggested step size if step is accepted
    def _acceptStepSize(self,s,h):
        self._accept = True
        s = np.min([s,self.MAX_S]) # dont let s be too big
        self.logs.append('step accepted with s = {:.2f}'.format(s))
        # if s much larger than 1, increase the step size
        if s > self.incrF:
            h_suggest = s*h
            self.logs.append('increasing step size to {}'.format(h_suggest))
            return h_suggest
        return h


    def evolve(self,u,t0,tf,h_init=None,**kwargs):
        """ 
        This function propagates an initial value (array) of u given at time t0
        until a final time tf is reached using a RK method for stiff PDEs 
        
        INPUTS:
            u : np.array
                initial value input 
            t0 : float
                initial time at which u is evaluated
            tf : float
                end time at which propagation stops
            h_init : float, optional
                starting step size for RK method (default of (tf-t0)/100 if not specified)
            store_data : bool, optional 
                value that determines whether to keep track of the propagation array u
                at each step of the RK method. Values stored in self.u and self.t
            store_freq : int, optional 
                store propagation data in self.t and self.u after every [store_freq] step is taken
        OUTPUTS:
            u : np.array 
                final value of the input propagated from t0 to tf 
        """
        
        self.reset()
        self.__t0, self.__tf, self.__tc = t0, tf, t0
        store_data = True
        if 'store_data' in kwargs:
             store_data = kwargs['store_data']
        store_freq = 1
        if 'store_freq' in kwargs:
            store_freq = kwargs['store_freq']
                   
        if store_data:
            self.t.append(t0)
            self.u.append(u)
        
        # Set initial step size if none given
        if h_init is None:
            h_init = (self.__tf-self.__t0)/100.    
        h = h_init
        
        # Make sure step size isn't larger than entire propagation time
        if self.__tc+h > self.__tf:
            h = self.__tf - self.__tc
        
        step_count = 0
        while self.__tc < self.__tf:
            u,h,h_suggest = self.step(u,h)
            self.__tc += h
            step_count += 1
            if self.__tc+h_suggest > self.__tf:
                h = self.__tf - self.__tc
            else:
                h = h_suggest
            
            if store_data and (step_count % store_freq == 0):
                self.t.append(self.__tc)
                self.u.append(u)
        
        return u








import numpy as np
from rkstiff.solver import StiffSolverAS
from scipy.linalg import expm

class IF45DP(StiffSolverAS):
    """
    Integrating factor adaptive step solver of 5th order with 4rd order embedding.
    Based on underlying Dormand-Prince method.

    ATTRIBUTES
    __________
    linop : np.array
    NLfunc : function
    t : time-array stored with evolve function call 
    u : output-array stored with evolve function call 
    logs : array of info stored related to the solver

    StiffSolverAS Parameters (see StiffSolverAS class in solver module)
    ________________________
    epsilon : float
    incrF : float
    decrF : float 
    safetyF : float
    adapt_cutoff : float
    minh : float 
    """
    def __init__(self,linop,NLfunc,**kwargs):
        """
        INPUTS
        ______

        linop : np.array
            Linear operator (L) in the system dtU = LU + NL(U). Can be either a 2D numpy array (matrix)
            or a 1D array (diagonal system). L can be either real-valued or complex-valued.

        NLfunc : function 
            Nonlinear function (NL(U)) in the system dtU = LU + NL(U). Can be a complex or real-valued function.

        diagonalize : bool, optional
            Diagonalize the linear operator (matrix) and solve the diagonalized system 

        StiffSolverAS variables: epsilon, incrF, decrF, safetyF, adapt_cutoff, minh 
                        (see StiffSolverAS documentation from solver module)
        """

        super().__init__(linop,NLfunc,**kwargs)
        if len(linop.shape) > 1:
            raise Exception('IF45DP only handles 1D linear operators (diagonal systems): try IF34,ETD34, or ETD35')
        self._EL15, self._EL310, self._EL45, self._EL89,self._EL = \
                [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(5)]
        self._NL1, self._NL2, self._NL3, self._NL4,self._NL5,self._NL6,\
                self._NL7 = [np.zeros(self.linop.shape[0],dtype=np.complex128) for _ in range(7)]
        self._a21, self._a31, self._a32, self._a41, self._a42, self._a43,\
                self._a51, self._a52, self._a53, self._a54 = [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(10)]
        self._a61, self._a62, self._a63, self._a64, self._a65 = [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(5)]
        self._a71, self._a73, self._a74, self._a75 = [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(4)]
        self._a76 = 0.0
        self._r1, self._r3, self._r5, self._r5 = [np.zeros(shape=self.linop.shape,dtype=np.complex128) for _ in range(4)]
        self._r6, self._r7 = 0.0, 0.0
        self._k = np.zeros(self.linop.shape[0],dtype=np.complex128)
        self._err = np.zeros(self.linop.shape[0],dtype=np.complex128)
        self._h_coeff = None
        self.__N1_init = False
        
    def _reset(self):
        self._h_coeff = None
        self.__N1_init = False
        
    def _updateStages(self,u,h):
        self._updateCoeffs(h)
        # First is same as last principle
        if not self.__N1_init:
            self._NL1 = self.NLfunc(u)
            self.__N1_init = True
        elif self._accept:
            self._NL1 = self._NL7.copy()
        
        self._k = self._EL15*u + self._a21*self._NL1
        self._NL2 = self.NLfunc(self._k)
        self._k = self._EL310*u + self._a31*self._NL1 + self._a32*self._NL2
        self._NL3 = self.NLfunc(self._k)
        self._k = self._EL45*u + self._a41*self._NL1 + self._a42*self._NL2 + self._a43*self._NL3
        self._NL4 = self.NLfunc(self._k)
        self._k = self._EL89*u + self._a51*self._NL1 + self._a52*self._NL2 + self._a53*self._NL3 \
                + self._a54*self._NL4
        self._NL5 = self.NLfunc(self._k)
        self._k = self._EL*u + self._a61*self._NL1 + self._a62*self._NL2 + self._a63*self._NL3 \
                + self._a64*self._NL4 + self._a65*self._NL5
        self._NL6 = self.NLfunc(self._k)  
        self._k = self._EL*u + self._a71*self._NL1 + self._a73*self._NL3 + self._a74*self._NL4 \
                + self._a75*self._NL5 + self._a76*self._NL6
        self._NL7 = self.NLfunc(self._k)
        self._err = self._r1*self._NL1 + self._r3*self._NL3 + self._r4*self._NL4 \
                  + self._r5*self._NL5  + self._r6*self._NL6 + self._r7*self._NL7

        return self._k, self._err
            
    def _updateCoeffs(self,h):
        # Update coefficients if step size h changed
        if h == self._h_coeff:
            return
        self._h_coeff = h
        z = h*self.linop
        self._EL15 = np.exp(z/5)
        self._EL310 = np.exp(3*z/10)
        self._EL45 = np.exp(4*z/5)
        self._EL89 = np.exp(8*z/9)
        self._EL = np.exp(z)
        EL710 = np.exp(7*z/10)
        EL19 = np.exp(z/9)
        self._a21 = h*self._EL15/5.0
        self._a31 = 3*h*self._EL310/40.0
        self._a32 = 9*h*np.exp(z/10)/40.0
        self._a41 = 44*h*self._EL45/45.0
        self._a42 = -56*h*np.exp(3*z/5)/15.0
        self._a43 = 32*h*np.exp(z/2)/9.0
        self._a51 = 19372*h*self._EL89/6561.0
        self._a52 = -25360*h*np.exp(31*z/45)/2187.0
        self._a53 = 64448.0*h*np.exp(53*z/90)/6561.0
        self._a54 = -212*h*np.exp(4*z/45)/729.0
        self._a61 = 9017*h*self._EL/3168.0
        self._a62 = -355*h*self._EL45/33.0
        self._a63 = 46732*h*EL710/5247.0
        self._a64 = 49*h*self._EL15/176.0
        self._a65 = -5103*h*EL19/18656.0
        self._a71 = 35*h*self._EL/384.0
        self._a73 = 500*h*EL710/1113.0
        self._a75 = -2187*h*EL19/6784.0
        self._a74 = 125*h*self._EL15/192.0
        self._a76 = 11*h/84.0
        self._r1 = h*71*self._EL/57600.0
        self._r3 = -71*h*EL710/16695.0
        self._r4 = 17*h*self._EL15/1920.0
        self._r5 = -17253*h*EL19/339200.0
        self._r6 = 22*h/525.0
        self._r7 = -h/40.0
           
        self.logs.append("IF45DP coefficients updated")
        
    def _q(self):
        return 5







import numpy as np

def dx_rfft(kx,u,n=1):
    """ Takes derivative(s) of a real valued array in spectral space 
    INPUTS
        kx - wavenumbers of the spectral grid
        u - real valued array
        n - order of derivative (default is 1)
    OUTPUTS
        uxn - derivative of u to the nth power
    """
    if not isinstance(n,int):
        raise TypeError('derivative order n must be an integer, it is {}'.format(n))
    if n < 0:
        raise ValueError('derivative order n must non-negative, it is {}'.format(n))

    if n == 0:
        return u

    uFFT = np.fft.rfft(u)
    if n == 1:
        uxn = np.fft.irfft(1j*kx*uFFT)
    else:
        uxn = np.fft.irfft(np.power(1j*kx,n)*uFFT)
    return uxn

def dx_fft(kx,u,n=1):
    """ Takes derivative(s) of a complex valued array in spectral space 
    INPUTS
        kx - wavenumbers of the spectral grid
        u - complex valued array
        n - order of derivative (default is 1)
    OUTPUTS
        uxn - derivative of u to the nth power
    """

    if not isinstance(n,int):
        raise TypeError('derivative order n must be an integer, it is {}'.format(n))
    if n < 0:
        raise ValueError('derivative order n must non-negative, it is {}'.format(n))
    if n == 0:
        return u

    uFFT = np.fft.fft(u)
    if n == 1:
        uxn = np.fft.ifft(1j*kx*uFFT)
    else:
        uxn = np.fft.ifft(np.power(1j*kx,n)*uFFT)
    return uxn






import numpy as np
import matplotlib.pyplot as plt

def orderTest(hvec,steps,solver,u0,uexact):
    errors = []
    for i,h in enumerate(hvec):
        u = u0.copy()
        for _ in range(steps[i]):
            u,hnew,h_suggest = solver.step(u,h)
            assert (hnew == h)
        errors.append(np.linalg.norm(u-uexact)/np.linalg.norm(uexact))
        solver.reset()
    return errors

def orderPlot(hvec,errors):
    
    m,c = np.polyfit(np.log(hvec),np.log(errors),1)
    pred = np.exp(m*np.log(hvec) + c)
    plt.plot(hvec,errors,'b',label='solver')
    plt.plot(hvec,pred,'r',label='fit of order {:.2f}'.format(m))
    plt.xlabel('h')
    plt.ylabel('error')
    plt.xscale('log')
    plt.yscale('log')
    plt.grid(True)
    plt.gca().invert_xaxis()
    plt.legend(loc='upper right')






from setuptools import setup

with open("README.md","r") as f:
    long_description = f.read()
exec(open("rkstiff/version.py").read())

setup(
        name='rkstiff',
        version=__version__,
        description='Runge-Kutta adaptive-step solvers for nonlinear PDEs',
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/whalenpt/rkstiff",
        author="Patrick Whalen",
        author_email="whalenpt@gmail.com",
        classifiers=[
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            ],
        extras_require= {
            "dev": [
                "pytest","twine",
                ],
            },
        packages=["rkstiff"],
        package_dir={'.' : 'rkstiff'},
        python_requires='>=3.6.0',
        setup_requires=["numpy>=1.14.0","scipy>=1.3.2"],
        install_requires=["numpy>=1.14.0","scipy>=1.3.2"]
)
